/*
 This file is part of the digitalSTROM dsm-api library.

 Copyright (C) 2010-2013 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM dsm-api library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 */

#include "dsm-api.h"

#include <stdlib.h>
#include <errno.h>
#include <pthread.h>

#include <utlist.h>
#include <uthash.h>

#include <digitalSTROM/ds485-client-interface.h>
#include "dsm-macros.h"

typedef struct {
    ds485ClientHandle_t client;
    struct callback_item *callback_list;
    BusState_callback_t bus_state_callback;
    void *bus_state_arg;
    BusChange_callback_t bus_change_callback;
    void *bus_change_arg;
    pthread_mutex_t mutex_callback;
    pthread_mutex_t mutex;
    pthread_cond_t condition;
    void *sync_data;
} dsm_api_data_t;

static int dsm_api_ds485_message_callback(void *handle, const ds485_container_t *container);
static int dsm_api_ds485_netlib_packet_callback(void *handle, const ds485n_packet_t *packet);
static int dsm_api_ds485_bus_state_callback(void *handle, const bus_state_t state);
static int dsm_api_ds485_bus_change_callback(void *handle, dsuid_t *id, int flags);
static int DsmApiInitializeProductLists();

typedef void (*callback_function_t)();

struct callback_item {
    uint32_t id;
    callback_function_t callback;
    void *arg;
    UT_hash_handle hh;
};

struct product_data {
    enum ptype {eDSD, eDSM, eDSS, eEthernet} product_type;

    uint16_t product_id;  // common
    uint16_t vendor_id;   // common
    uint8_t group_id;     // only: eDSD
    char *ieee_prefix;    // only: eEthernet
    uint8_t ieee_prefix_length; // only: eEthernet

    char *hardware_info;
    char *display_name;
    char *gtin;
    dsid_t sgtin_prefix;
    struct product_data* next;
};

struct product_item {
    struct product_data* pdata;
    uint64_t id;
    UT_hash_handle hh;
};

static pthread_mutex_t product_list_mutex = PTHREAD_MUTEX_INITIALIZER;
static struct product_data* dsd_product_list = NULL;
static struct product_data* dsm_product_list = NULL;
static struct product_data* eth_product_list = NULL;
static struct product_item* product_hash = NULL;


int DsmApiLibraryVersion(void)
{
    return API_VERSION;
}

DsmApiHandle_t DsmApiInitialize(void)
{
    int err;

    dsm_api_data_t *data = calloc(1, sizeof(dsm_api_data_t));
    if (data == NULL) {
        return NULL;
    }

    pthread_mutex_init(&data->mutex_callback, NULL);
    pthread_mutex_init(&data->mutex, NULL);
    pthread_cond_init(&data->condition, NULL);

    err = DsmApiInitializeProductLists();
    if (err != 0) {
        return NULL;
    }

    return data;
}

int DsmApiOpen(DsmApiHandle_t handle, const char *connection, int flags)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    struct ds485c_callbacks g_cb = {
        .state_change_cb = dsm_api_ds485_bus_state_callback,
        .state_change_data = data,
        .bus_change_cb = dsm_api_ds485_bus_change_callback,
        .bus_change_data = data,
        .container_pkt_cb = dsm_api_ds485_message_callback,
        .container_pkt_data = data,
        .netlib_pkt_cb = dsm_api_ds485_netlib_packet_callback,
        .netlib_pkt_data = data,
    };

    data->client = ds485_client_open2(connection, flags, &g_cb);
    if (data->client == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    return 0;
}

int DsmApiClose(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    struct callback_item *item;

    pthread_mutex_lock(&data->mutex_callback);
    while (data->callback_list)
    {
        item = data->callback_list;
        HASH_DEL(data->callback_list, item);
        free(item);
    }
    pthread_mutex_unlock(&data->mutex_callback);

    /* CAUTION: possible callback in parallel */
    int result = ds485_client_close(data->client);

    data->client = NULL;

    return result;
}

int DsmApiCleanup(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL || data->client != NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    pthread_mutex_destroy(&data->mutex_callback);
    pthread_mutex_destroy(&data->mutex);
    pthread_cond_destroy(&data->condition);

    free(data);

    return 0;
}

int DsmApiSetCallback(DsmApiHandle_t handle, container_type_t MessageType, uint8_t MessageId, uint8_t ModifierId, DsmApiCallback_t *callback_struct)
{
    struct callback_item *callback;

    uint32_t id = (MessageType << 16) | (MessageId << 8) | (ModifierId);

    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    pthread_mutex_lock(&data->mutex_callback);
    HASH_FIND_INT(data->callback_list, &id, callback);

    if (callback == NULL)
    {
        /* callback not yet in hash */
        if (callback_struct->function != NULL)
        {
            /* create new entry in hash */
            callback = (struct callback_item *)malloc(sizeof(struct callback_item));

            callback->id = id;
            callback->callback = (callback_function_t)callback_struct->function;
            callback->arg = callback_struct->arg;
            HASH_ADD_INT(data->callback_list, id, callback);
            callback_struct->function = NULL;
            callback_struct->arg = NULL;
        }
    }
    else
    {
        /* found callback in hash */
        if (callback_struct->function != NULL)
        {
            DsmApiCallback_t prev_callback;
            prev_callback.function = callback->callback;
            prev_callback.arg = callback->arg;
            /* replace previous entry in hash */
            callback->callback = (callback_function_t)callback_struct->function;
            callback->arg = callback_struct->arg;
            /* no need to add, as it's already part of the hash */
            /* copy previous callback back to user */
            *callback_struct = prev_callback;
        }
        else
        {
            HASH_DEL(data->callback_list, callback);
            /* copy previous callback back to user */
            callback_struct->function = callback->callback;
            callback_struct->arg = callback->arg;
            free(callback);
        }
    }
    pthread_mutex_unlock(&data->mutex_callback);

    return ERROR_OK;
}

int DsmApiSetBusStateCallback(DsmApiHandle_t handle, DsmApiCallback_t *callback_struct)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;
    DsmApiCallback_t prev_cb;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    prev_cb.function = data->bus_state_callback;
    prev_cb.arg = data->bus_state_arg;
    data->bus_state_callback = (BusState_callback_t)callback_struct->function;
    data->bus_state_arg = callback_struct->arg;
    *callback_struct = prev_cb;
    return ERROR_OK;
}

int DsmApiSetBusChangeCallback(DsmApiHandle_t handle, DsmApiCallback_t *callback_struct)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;
    DsmApiCallback_t prev_cb;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    prev_cb.function = data->bus_change_callback;
    prev_cb.arg = data->bus_change_arg;
    data->bus_change_callback = (BusChange_callback_t)callback_struct->function;
    data->bus_change_arg = callback_struct->arg;

    // copy previous struct back to caller
    *callback_struct = prev_cb;
    return ERROR_OK;
}

int DsmApiGetBusMembers(DsmApiHandle_t handle, dsuid_t *device_list, int device_list_size)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    return ds485_client_query_devices(data->client, device_list, device_list_size);
}

int DsmApiGetBusMembersPerType(DsmApiHandle_t handle, dsuid_t *device_list, int device_list_size, uint16_t types)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    dsuid_t tmp_device_list[63];
    int list_count = ds485_client_query_devices(data->client, tmp_device_list, 63);
    if (list_count < 0) {
        return list_count;
    }

    dsuid_t ownDsuid;
    if (DsmApiGetOwnDSUID(handle, &ownDsuid) != ERROR_OK) {
        SetNullDsuid(ownDsuid);
    }
    uint8_t i;
    int list_size = 0;
    for (i = 0; i < list_count; i++) {
        uint8_t busMemberType;
        if (IsEqualDsuid(tmp_device_list[i], ownDsuid)) {
            busMemberType = BusMember_ds485d;
        } else {
            if (BusMember_get_type(handle, tmp_device_list[i], &busMemberType) != ERROR_OK) {
                continue;
            }
        }

        if (list_size >= device_list_size) {
            break;
        }
        if (((types & BUS_MEMBER_GET_DS485D) && (busMemberType == BusMember_ds485d)) ||
            ((types & BUS_MEMBER_GET_DS485P) && (busMemberType == BusMember_ds485p)) ||
            ((types & BUS_MEMBER_GET_DSM11) && (busMemberType == BusMember_dSM11)) ||
            ((types & BUS_MEMBER_GET_DSM12) && (busMemberType == BusMember_dSM12)) ||
            ((types & BUS_MEMBER_GET_VDSM) && (busMemberType == BusMember_vDSM)) ||
            ((types & BUS_MEMBER_GET_VDC) && (busMemberType == BusMember_vDC)) ||
            ((types & BUS_MEMBER_GET_DSS) && (busMemberType == BusMember_dSS)) ||
            ((types & BUS_MEMBER_GET_DSA) && (busMemberType == BusMember_dSA)) ||
            ((types & BUS_MEMBER_GET_DSG) && (busMemberType == BusMember_dSG)))
        {
            device_list[list_size++] = tmp_device_list[i];
        }
    }
    return list_size;
}

int DsmApiGetBusState(DsmApiHandle_t handle, bus_state_t *state)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    return ds485_client_query_state(data->client, state);
}

int DsmApiRestartBus(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    return ds485_client_restart_bus(data->client);
}

static int dsm_api_ds485_bus_state_callback(void *handle, const bus_state_t state)
{
    dsm_api_data_t *data = handle;

    if (data == NULL || data->bus_state_callback == NULL) {
        return 0;
    }

    data->bus_state_callback(data->bus_state_arg, state);
    return 0;
 }

static int dsm_api_ds485_bus_change_callback(void *handle, dsuid_t *id, int flags)
{
    dsm_api_data_t *data = handle;

    if (data == NULL || data->bus_change_callback == NULL) {
        return 0;
    }

    data->bus_change_callback(data->bus_change_arg, id, flags);
    return 0;
}

int DsmApiGetOwnDSUID(DsmApiHandle_t handle, dsuid_t *dsuid)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    return ds485_client_get_dsuid(data->client, dsuid);
}

int DsmApiGetDeviceDSUID(uint16_t vendorId, uint16_t productId, uint8_t groupId, uint16_t revisionId, uint32_t serialNumber, dsuid_t *dsuid)
{
    static const dsuid_t aizoDevice = {0x35, 0x04, 0x17, 0x5f, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
    int retval = ERROR_DSUID_CONVERSION;

    pthread_mutex_lock(&product_list_mutex);
    if (dsd_product_list == NULL)
    {
        pthread_mutex_unlock(&product_list_mutex);
        DsmApiInitializeProductLists();
        pthread_mutex_lock(&product_list_mutex);
    }

    memset(dsuid->id, 0, sizeof(dsuid_t));

    // GID-96 schema: "old" devices should not change their dSUID
    if ((revisionId < 0x0355) || ((revisionId >= 0xA000) && (vendorId == 1)))
    {
        *dsuid = aizoDevice;
        retval = ERROR_OK;
    }

    // SGTIN-96 schema
    else
    {
        struct product_item *pitem = NULL;
        struct product_data *pdata = NULL;
        uint64_t id =
                (uint64_t) vendorId |
                ((uint64_t) productId << 16) |
                ((uint64_t) groupId << 32) |
                ((uint64_t) eDSD << 48);

        HASH_FIND(hh, product_hash, &id, sizeof(uint64_t), pitem);
        if (pitem == NULL)
        {
            LL_FOREACH(dsd_product_list, pdata)
            {
                if (pdata->vendor_id == vendorId &&
                    pdata->product_id == productId &&
                    pdata->group_id == groupId)
                {
                    break;
                }
            }
            if (pdata)
            {
                pitem = (struct product_item*) malloc(sizeof(struct product_item));
                pitem->pdata = pdata;
                pitem->id = id;
                HASH_ADD(hh, product_hash, id, sizeof(uint64_t), pitem);
            }
        }
        if (pitem)
        {
            memcpy(&dsuid->id[0], &pitem->pdata->sgtin_prefix.id[0], 6);
            memcpy(&dsuid->id[10], &pitem->pdata->sgtin_prefix.id[6], 6);
            retval = ERROR_OK;
        }
    }

    pthread_mutex_unlock(&product_list_mutex);

    dsuid->id[12] = (serialNumber >> 24) & 0xFF;
    dsuid->id[13] = (serialNumber >> 16) & 0xFF;
    dsuid->id[14] = (serialNumber >> 8) & 0xFF;
    dsuid->id[15] = (serialNumber >> 0) & 0xFF;

    return retval;
}

int DsmApiAddDeviceDSUIDHeader(uint16_t vendorId, uint16_t productId, uint8_t groupId, uint16_t revisionId, dsuid_t *dsuid)
{
    dsuid_t tmp_dsuid;
    int retval;
    uint32_t serialNumber;

    if ((dsuid->id[6] != 0) ||
        (dsuid->id[7] != 0) ||
        (dsuid->id[8] != 0) ||
        (dsuid->id[9] != 0))
    {
        return ERROR_WRONG_PARAMETER;
    }

    serialNumber = ((uint32_t)dsuid->id[12] << 24) |
                   ((uint32_t)dsuid->id[13] << 16) |
                   ((uint32_t)dsuid->id[14] << 8) |
                   (uint32_t)dsuid->id[15];

    retval = DsmApiGetDeviceDSUID(vendorId, productId, groupId, revisionId, serialNumber, &tmp_dsuid);

    if (retval == ERROR_OK) {
        memcpy(dsuid, &tmp_dsuid, sizeof(dsuid_t));
    }

    return retval;
}

int DsmApiGetEthernetDSUID(const char *mac, dsuid_t *dsuid)
{
    int retval = ERROR_DSUID_CONVERSION;
    char *p;
    int i;
    uint32_t vendor = 0;
    uint32_t serial = 0;
    char mac_string[13];
    struct product_data *pdata = NULL;

    p = (char *) mac;
    mac_string[12] = 0;
    for (i = 0; i < 3; i++)
    {
        char number[3] =
            { 0, };
        memcpy(number, p, 2);
        uint8_t n = strtol(number, NULL, 16);
        vendor = (vendor << 8) | n;
        memcpy(mac_string + i*2, p, 2);
        p += 2;
        if ((*p == ':') || (*p == '-'))
        {
            p++;
        }
    }
    for (i = 0; i < 3; i++)
    {
        char number[3] =
            { 0, };
        memcpy(number, p, 2);
        uint8_t n = strtol(number, NULL, 16);
        serial = (serial << 8) | n;
        memcpy(mac_string + (i+3)*2, p, 2);
        p += 2;
        if ((*p == ':') || (*p == '-'))
        {
            p++;
        }
    }

    pthread_mutex_lock(&product_list_mutex);
    if (eth_product_list == NULL)
    {
        pthread_mutex_unlock(&product_list_mutex);
        DsmApiInitializeProductLists();
        pthread_mutex_lock(&product_list_mutex);
    }
    LL_FOREACH(eth_product_list, pdata)
    {
        if (strncasecmp(pdata->ieee_prefix, mac_string, pdata->ieee_prefix_length) == 0)
        {
            break;
        }
    }
    pthread_mutex_unlock(&product_list_mutex);

    // SGTIN-96 schema
    if (pdata)
    {
        retval = 0;
        memset(dsuid, 0, sizeof(dsuid_t));
        memcpy(&dsuid->id[0], &pdata->sgtin_prefix.id[0], 6);
        memcpy(&dsuid->id[10], &pdata->sgtin_prefix.id[6], 6);
        // serial number bits 0..37, insert the 14 bits product-id prefix starting from leftmost MSB
        uint16_t pid = (pdata->product_id & 0x3FFF);
        // swap odd and even bits
        pid = ((pid >> 1) & 0x5555) | ((pid & 0x5555) << 1);
        // swap consecutive pairs
        pid = ((pid >> 2) & 0x3333) | ((pid & 0x3333) << 2);
        // swap nibbles ...
        pid = ((pid >> 4) & 0x0F0F) | ((pid & 0x0F0F) << 4);
        // swap bytes
        pid = ((pid >> 8) & 0x00FF) | ((pid & 0x00FF) << 8);
        // shift down as the two MSB's are zero anyway
        pid >>= 2;
        dsuid->id[11] &= 0xC0;
        dsuid->id[11] |= (pid >> 8);
        dsuid->id[12] = (pid & 0xFF);
    }

    // fallback to GID-96 scheme with aizo Manager Number
    else
    {
        const dsuid_t aizoDevice = {0x35, 0x04, 0x17, 0x5f, 0xef, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
        *dsuid = aizoDevice;
        dsuid->id[5] = 0xf0 | ((vendor >> 20) & 0xf);
        dsuid->id[10] = (vendor >> 12) & 0xff;
        dsuid->id[11] = (vendor >> 4) & 0xf0;  // NOTE: lower 4 bits zero
        dsuid->id[12] = vendor & 0xff;
    }

    dsuid->id[13] = (serial >> 16) & 0xFF;
    dsuid->id[14] = (serial >> 8) & 0xFF;
    dsuid->id[15] = (serial >> 0) & 0xFF;

    return retval;
}

int DsmApiGetDeviceDescription(uint16_t vendorId, uint16_t productId, uint8_t groupId, uint16_t revisionId,
    char **displayName, char **hardwareInfo, char **gtin)
{
    struct product_item *pitem = NULL;
    struct product_data *pdata = NULL;
    uint64_t id =
            (uint64_t) vendorId |
            ((uint64_t) productId << 16) |
            ((uint64_t) groupId << 32) |
            ((uint64_t) eDSD << 48);

    pthread_mutex_lock(&product_list_mutex);
    HASH_FIND_INT(product_hash, &id, pitem);
    if (!pitem)
    {
        LL_FOREACH(dsd_product_list, pdata)
        {
            if (pdata->vendor_id == vendorId &&
                pdata->product_id == productId &&
                pdata->group_id == groupId)
            {
                break;
            }
        }
        if (pdata)
        {
            pitem = (struct product_item*) malloc(sizeof(struct product_item));
            pitem->pdata = pdata;
            pitem->id = id;
            HASH_ADD(hh, product_hash, id, sizeof(uint64_t), pitem);
        }
    }
    pthread_mutex_unlock(&product_list_mutex);

    if (pitem)
    {
        if (displayName)
        {
            *displayName = pitem->pdata->display_name;
        }
        if (hardwareInfo)
        {
            *hardwareInfo = pitem->pdata->hardware_info;
        }
        if (gtin)
        {
            *gtin = pitem->pdata->gtin;
        }
        return ERROR_OK;
    }
    return ERROR_DSUID_CONVERSION;
}

int DsmApiIsdSM(dsuid_t dSMdSUID)
{
    struct product_data *pdata = NULL;

    // GID-96 schema
    if (dSMdSUID.id[0] == 0x35)
    {
        if (dSMdSUID.id[1] == 0x04 && dSMdSUID.id[2] == 0x17 &&
            dSMdSUID.id[3] == 0x5f && dSMdSUID.id[4] == 0xe0 &&
            dSMdSUID.id[5] == 0x00 && dSMdSUID.id[10] == 0x00 &&
            ((dSMdSUID.id[11] & 0xf0) == 0x10))
        {
            return 1;
        }
        return 0;
    }

    dSMdSUID.id[11] &= 0xC0;

    pthread_mutex_lock(&product_list_mutex);
    LL_FOREACH(dsm_product_list, pdata)
    {
        if ((memcmp(dSMdSUID.id, pdata->sgtin_prefix.id, 6) == 0) &&
            (dSMdSUID.id[10] == pdata->sgtin_prefix.id[6]) &&
            (dSMdSUID.id[11] == pdata->sgtin_prefix.id[7]))
        {
            break;
        }
    }
    pthread_mutex_unlock(&product_list_mutex);

    if (pdata && pdata->product_type == eDSM)
    {
        return 1;
    }

    return 0;
}



typedef enum {
    IDLE,
    WAITING,
    DONE,
} CallbackSyncState_t;

typedef struct {
    CallbackSyncState_t state;
    dsuid_t dsmId;
    uint16_t device;
    uint8_t offset;
    uint8_t bank;
    uint8_t scene;
    uint16_t result;
    uint8_t result2;
} CallbackSync_t;

static void SyncDeviceStatistics_handler(uint8_t error_code, void *arg, dsuid_t sourceId, dsuid_t destinationId, uint16_t DeviceId, uint8_t Offset, uint16_t Value)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSync_t *data = (CallbackSync_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            IsEqualDsuid(data->dsmId, sourceId) &&
            data->device == DeviceId &&
            data->offset == Offset)
        {
            data->result = Value;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceStatisticsRequest_get_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ParamOffset, int timeout, uint16_t *Value)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceStatistics_event_callback_t response_callback = &SyncDeviceStatistics_handler;
    void *previousCallback;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    if (ParamOffset == 0xfe || ParamOffset == 0xff)
    {
        return ERROR_INVALID_PARAMETER;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;
    data->offset = ParamOffset;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_STATISTICS, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceStatisticsRequest_get(handle, dSMdSUID, DeviceId, ParamOffset);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *Value = data->result;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_STATISTICS, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

int DeviceConfig_get_sync_8(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint8_t *Value)
{
    uint16_t value16;
    int result = DeviceConfig_get_sync_16(handle, dSMdSUID, DeviceId, ConfigClass, (ConfigIndex & 0xFE), timeout, &value16);
    *Value = ((ConfigIndex & 0x1) == 0) ? (value16 & 0xFF) : (value16 >> 8);
    return result;
}

static void SyncDeviceConfig_16_handler(uint8_t error_code, void *arg, dsuid_t sourceId, dsuid_t destinationId, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, uint8_t Value)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSync_t *data = (CallbackSync_t*)apiHandle->sync_data;
    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            IsEqualDsuid(data->dsmId, sourceId) &&
            data->device == DeviceId &&
            data->bank == ConfigClass)
        {
            if (data->offset == ConfigIndex)
            {
                data->result += Value;
                data->result2 |= 0x01;
            }
            else if (data->offset + 1 == ConfigIndex)
            {
                data->result += (uint16_t)Value << 8;
                data->result2 |= 0x02;
            }
            if (data->result2 == 0x03)
            {
                data->state = DONE;
                pthread_cond_signal(&apiHandle->condition);
            }
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceConfig_get_sync_16(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint16_t *Value)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceConfig_event_callback_t response_callback = &SyncDeviceConfig_16_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    if ((ConfigIndex & 0x1) != 0)
    {
        return ERROR_INVALID_PARAMETER;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;
    data->offset = ConfigIndex & 0xFE;
    data->bank = ConfigClass;
    data->result = 0;
    data->result2 = 0;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceConfig_get(handle, dSMdSUID, DeviceId, ConfigClass, ConfigIndex);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *Value = data->result;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

int DeviceOPCConfig_get_current_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ChannelId, int timeout, uint16_t *Value)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceConfig_event_callback_t response_callback = &SyncDeviceConfig_16_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;
    data->offset = 2 * ChannelId;
    data->bank = 100;
    data->scene = 0;
    data->result = 0;
    data->result2 = 0;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceOPCConfig_get_current(handle, dSMdSUID, DeviceId, ChannelId);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *Value = data->result;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

int DeviceOPCConfig_get_scene_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t ChannelId, uint8_t SceneId, int timeout, uint16_t *Value)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceConfig_event_callback_t response_callback = &SyncDeviceConfig_16_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;
    data->offset = 2 * ChannelId;
    data->bank = 102;
    data->scene = SceneId;
    data->result = 0;
    data->result2 = 0;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceOPCConfig_get_scene(handle, dSMdSUID, DeviceId, ChannelId, SceneId);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *Value = data->result;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

int DeviceOPCConfig_get_scene_config_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t SceneId, int timeout, uint16_t *Value)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceConfig_event_callback_t response_callback = &SyncDeviceConfig_16_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;
    data->offset = 2 * SceneId;
    data->bank = 103;
    data->scene = SceneId;
    data->result = 0;
    data->result2 = 0;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceOPCConfig_get_scene_config(handle, dSMdSUID, DeviceId, SceneId);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *Value = data->result;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

int DeviceOPCConfig_get_dc_flags_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t SceneId, int timeout, uint16_t *Value)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceConfig_event_callback_t response_callback = &SyncDeviceConfig_16_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;
    data->offset = 2 * SceneId;
    data->bank = 104;
    data->scene = SceneId;
    data->result = 0;
    data->result2 = 0;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceOPCConfig_get_dc_flags(handle, dSMdSUID, DeviceId, SceneId);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *Value = data->result;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

static void SyncTestTransmissionQuality_handler(uint8_t error_code, void *arg, dsuid_t sourceId, dsuid_t destinationId, uint16_t DeviceId, uint8_t DownstreamQuality, uint16_t UpstreamQuality)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSync_t *data = (CallbackSync_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            IsEqualDsuid(data->dsmId, sourceId) &&
            data->device == DeviceId)
        {
            data->result = UpstreamQuality;
            data->result2 = DownstreamQuality;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int TestTransmissionQuality_get_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, int timeout, uint8_t *DownstreamQuality, uint16_t *UpstreamQuality)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventTestTransmissionQuality_event_callback_t response_callback = &SyncTestTransmissionQuality_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_TEST_TRANSMISSION_QUALITY, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = TestTransmissionQuality_get(handle, dSMdSUID, DeviceId);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *UpstreamQuality = data->result;
            *DownstreamQuality = data->result2;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_TEST_TRANSMISSION_QUALITY, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

typedef struct {
    CallbackSyncState_t state;
    dsuid_t dsmId;
    uint16_t device;
    uint8_t SensorIndex;
    uint16_t SensorValue;
} CallbackSyncSensorValue_t;

static void SyncDeviceSensor_get_handler(uint8_t error_code, void *arg, dsuid_t sourceId, dsuid_t destinationId, uint16_t DeviceId, uint8_t SensorIndex, uint16_t SensorValue)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSyncSensorValue_t *data = (CallbackSyncSensorValue_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            IsEqualDsuid(data->dsmId, sourceId) &&
            data->device == DeviceId &&
            data->SensorIndex == SensorIndex)
        {
            data->SensorValue = SensorValue;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceSensor_get_value_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t SensorIndex, int timeout, uint16_t *SensorValue)
{
    CallbackSyncSensorValue_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceSensor_value_event_callback_t response_callback = &SyncDeviceSensor_get_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSyncSensorValue_t*)malloc(sizeof(CallbackSyncSensorValue_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;
    data->SensorIndex = SensorIndex;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_SENSOR, EVENT_DEVICE_SENSOR_VALUE, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceSensor_get_value(handle, dSMdSUID, DeviceId, SensorIndex);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *SensorValue = data->SensorValue;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_SENSOR, EVENT_DEVICE_SENSOR_VALUE, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

static void SyncDSLink_response_handler(uint8_t error_code, void *arg, dsuid_t sourceId, dsuid_t destinationId, uint16_t DeviceId, uint8_t Value)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSync_t *data = (CallbackSync_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            IsEqualDsuid(data->dsmId, sourceId) &&
            data->device == DeviceId)
        {
            data->result = Value;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceDslinkSend_last_byte_with_answer_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t Value, int timeout, uint8_t *ReturnValue)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDslink_response_event_callback_t response_callback = &SyncDSLink_response_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceDslinkSend_last_byte_with_answer(handle, dSMdSUID, DeviceId, Value);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *ReturnValue = (uint8_t)data->result;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

int DeviceDslinkSend_not_last_byte_with_answer_sync(DsmApiHandle_t handle, dsuid_t dSMdSUID, uint16_t DeviceId, uint8_t Value, int timeout, uint8_t *ReturnValue)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDslink_response_event_callback_t response_callback = &SyncDSLink_response_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->dsmId = dSMdSUID;
    data->device = DeviceId;

    callback.function = (void*)response_callback;
    callback.arg = data;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceDslinkSend_not_last_byte_with_answer(handle, dSMdSUID, DeviceId, Value);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        int rval = 0;
        while ((rval == 0) && (data->state != DONE)) {
            rval = pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *ReturnValue = (uint8_t)data->result;
            result = 0;
        } else {
            result = ERROR_SYNC_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

static char *DsmApiCVSgetline(FILE *f);
static char *DsmApiCVSfield(int n);
static int   DsmApiCVSfields(void);

struct db_parser_data
{
    char *line;
    char *sline;
    int maxline;
    char **field;
    int maxfield;
    int nfield;
    char* fieldsep;
};

static struct db_parser_data db_parser =
    { NULL, NULL, 0, NULL, 0, 0, ";," };

static int DsmApiCVSendofline(FILE *fin, int c)
{
    int eol = (c == '\r' || c == '\n');
    if (c == '\r')
    {
        c = getc(fin);
        if (c != '\n' && c != EOF)
        {
            ungetc(c, fin);
        }
    }
    return eol;
}

static void DsmApiCVSreset(void)
{
    free(db_parser.line);
    free(db_parser.sline);
    free(db_parser.field);
    db_parser.line = NULL;
    db_parser.sline = NULL;
    db_parser.field = NULL;
    db_parser.maxline = db_parser.maxfield = db_parser.nfield = 0;
}

static char* DsmApiCVSquotes(char *p)
{
    int i, j;

    for (i = j = 0; p[j] != '\0'; i++, j++)
    {
        if (p[j] == '"' && p[++j] != '"')
        {
            int k = strcspn(p + j, db_parser.fieldsep);
            memmove(p + i, p + j, k);
            i += k;
            j += k;
            break;
        }
        p[i] = p[j];
    }
    p[i] = '\0';
    return p + j;
}

static int DsmApiCVSsplit(void)
{
    char *p, **newf;
    char *sepp;
    int sepc;

    db_parser.nfield = 0;
    if (db_parser.line[0] == '\0')
    {
        return 0;
    }

    strcpy(db_parser.sline, db_parser.line);
    p = db_parser.sline;
    do
    {
        if (db_parser.nfield >= db_parser.maxfield)
        {
            db_parser.maxfield *= 2;
            newf = (char **) realloc(db_parser.field,
                    db_parser.maxfield * sizeof(db_parser.field[0]));
            if (newf == NULL )
            {
                return -2;
            }
            db_parser.field = newf;
        }
        if (*p == '"')
        {
            sepp = DsmApiCVSquotes(++p);
        }
        else
        {
            sepp = p + strcspn(p, db_parser.fieldsep);
        }
        sepc = sepp[0];
        sepp[0] = '\0';
        db_parser.field[db_parser.nfield++] = p;
        p = sepp + 1;
    }
    while (sepc == ';' || sepc == ',');

    return db_parser.nfield;
}

static char* DsmApiCVSgetline(FILE *fin)
{
    int i, c;
    char *newl, *news;

    if (db_parser.line == NULL )
    {
        db_parser.maxline = db_parser.maxfield = 1;
        db_parser.line = (char *) malloc(db_parser.maxline);
        db_parser.sline = (char *) malloc(db_parser.maxline);
        db_parser.field = (char **) malloc(
                db_parser.maxfield * sizeof(db_parser.field[0]));
        if (db_parser.line == NULL || db_parser.sline == NULL
                || db_parser.field == NULL )
        {
            DsmApiCVSreset();
            return NULL ;
        }
    }
    for (i = 0; (c = getc(fin)) != EOF && !DsmApiCVSendofline(fin, c); i++)
    {
        if (i >= db_parser.maxline - 1)
        {
            db_parser.maxline *= 2;
            newl = (char *) realloc(db_parser.line, db_parser.maxline);
            if (newl == NULL )
            {
                DsmApiCVSreset();
                return NULL ;
            }
            db_parser.line = newl;
            news = (char *) realloc(db_parser.sline, db_parser.maxline);
            if (news == NULL )
            {
                DsmApiCVSreset();
                return NULL ;
            }
            db_parser.sline = news;
        }
        db_parser.line[i] = c;
    }
    db_parser.line[i] = '\0';
    if (DsmApiCVSsplit() == -2)
    {
        DsmApiCVSreset();
        return NULL ;
    }
    return (c == EOF && i == 0) ? NULL : db_parser.line;
}

static char* DsmApiCVSfield(int n)
{
    if (n < 0 || n >= db_parser.nfield)
    {
        return NULL ;
    }
    return db_parser.field[n];
}

static int DsmApiCVSfields(void)
{
    return db_parser.nfield;
}

static int DsmApiInitializeProductType(int ptype)
{
    FILE* fin;
    char *p = NULL;
    char *image_file = NULL;
    char *line = NULL;
    char *basename = NULL;
    char fname[1023];
    int i;
    struct product_data **plist;
    struct product_data *pdata;
    struct product_item *pitem;
    uint64_t id;

    switch (ptype)
    {
    // TODO: adapt dsdata to dsuid
    case eDSD:
        p = getenv("DSUID_DSD_TABLE");
        basename = "productlist_dsd.csv";
        plist = &dsd_product_list;
        break;
    case eDSM:
        p = getenv("DSUID_DSM_TABLE");
        basename = "productlist_dsm.csv";
        plist = &dsm_product_list;
        break;
    case eEthernet:
        p = getenv("DSUID_ETH_TABLE");
        image_file = "/etc/hardware_gtin.csv";
        basename = "productlist_eth.csv";
        plist = &eth_product_list;
        break;
    default:
        return -1;
    }

    if (*plist != NULL)
    {
        return 0;
    }
    if (image_file != NULL && (access(image_file, R_OK) == 0))
    {
        strcpy(fname, image_file);
    }
    else if (p != NULL && (access(p, R_OK) == 0))
    {
        strcpy(fname, p);
    }
    else
    {
        char *paths[] = {
                "./",
                "/usr/share/digitalstrom/",
                "/usr/local/share/digitalstrom/",
                "/etc/digitalstrom/"
        };
        for (i = 0; i < sizeof(paths) / sizeof(char *); i++)
        {
            strcpy(fname, paths[i]);
            strcat(fname, basename);
            if (access(fname, R_OK) == 0)
            {
                goto found;
            }
            fname[0] = 0;
        }

        p = getenv("DSDATA_PATH");
        if (p != NULL && (access(p, R_OK|X_OK) == 0)) {
            strcpy(fname, p);
            strcat(fname, "/");
            strcat(fname, basename);
            if (access(fname, R_OK) == 0) {
                goto found;
            }
        }
        fprintf(stderr, "ERROR: %s not found in standard directories; "
                        "please install dsdata or set DSDATA_PATH "
                        "environment variable correct\n", basename);
        return -ENOENT;
    }
found:
    if ((fin = fopen(fname, "rt")) == NULL )
    {
        return -ENOENT;
    }

    while ((line = DsmApiCVSgetline(fin)) != NULL )
    {
        if (line[0] == '#' || line[0] == ';' || line[0] == ',')
        {
            // ignore header, comment or empty lines
            continue;
        }
        if ((DsmApiCVSfields() < 6)
                || (DsmApiCVSfield(0)[0] == 0)
                || (DsmApiCVSfield(3)[0] == 0)
                || (DsmApiCVSfield(4)[0] == 0))
        {
            // ignore incomplete entries - need at least product names, gtin & sgtin data
            continue;
        }

        pdata = (struct product_data*) calloc(1, sizeof(struct product_data));
        pdata->product_type = ptype;
        pdata->display_name = strdup(DsmApiCVSfield(0));
        pdata->hardware_info = strdup(DsmApiCVSfield(1));
        pdata->gtin = strdup(DsmApiCVSfield(3));
        p = DsmApiCVSfield(4);
        memset(&pdata->sgtin_prefix, 0, sizeof(dsid_t));
        for (i = 0; i < strlen(p) && (i < 2 * sizeof(dsid_t)); i++)
        {
            char number[3] = { 0, };
            number[0] = p[i];
            if (i % 2 == 0)
            {
                pdata->sgtin_prefix.id[i/2] |= (strtol(number, NULL, 16) << 4);
            }
            else
            {
                pdata->sgtin_prefix.id[i/2] |= strtol(number, NULL, 16);
            }
        }
#undef DUMP_TABLES
#ifdef DUMP_TABLES
        printf("PRODUCT: type=%-2d name=%-14s hwinfo=%-10s gtin=%s, sgtin=",
                ptype, pdata->display_name, pdata->hardware_info, pdata->gtin);
        for (i = 0; i < sizeof(dsid_t); i++) {
            printf("%02X", pdata->sgtin_prefix.id[i]);
        }
#endif

        switch (ptype)
        {
            case eDSD:
                pdata->group_id = DsmApiCVSfield(5) ? strtol(DsmApiCVSfield(5), NULL, 0) : 0;
                pdata->product_id = DsmApiCVSfield(6) ? strtol(DsmApiCVSfield(6), NULL, 0) : 0;
                pdata->vendor_id = DsmApiCVSfield(7) ? strtol(DsmApiCVSfield(7), NULL, 0) : 0;
#ifdef DUMP_TABLES
                printf(" DSD Group 0x%02x, Product 0x%04x, Vendor 0x%04x\n",
                       pdata->group_id, pdata->product_id, pdata->vendor_id);
#endif
                break;
            case eDSM:
                pdata->product_id = DsmApiCVSfield(5) ? strtol(DsmApiCVSfield(5), NULL, 0) : 0;
                pdata->vendor_id = DsmApiCVSfield(6) ? strtol(DsmApiCVSfield(6), NULL, 0) : 0;
#ifdef DUMP_TABLES
                printf(" DSM Product 0x%04x, Vendor 0x%04x\n",
                       pdata->product_id, pdata->vendor_id);
#endif
                break;
            case eEthernet:
                pdata->product_id = DsmApiCVSfield(5) ? strtol(DsmApiCVSfield(5), NULL, 0) : 0;
                pdata->ieee_prefix = DsmApiCVSfield(6) ? strdup(DsmApiCVSfield(6)) : "";
                pdata->ieee_prefix_length = DsmApiCVSfield(7) ? strtol(DsmApiCVSfield(7), NULL, 0) : 6;
#ifdef DUMP_TABLES
                printf(" ETH Product 0x%04x, MAC %s\n",
                       pdata->product_id, pdata->ieee_prefix);
#endif
                break;
        }

        LL_APPEND(*plist, pdata);
    }
    fclose(fin);
    return 0;
}

static int DsmApiInitializeProductLists()
{
    int err = 0;

    pthread_mutex_lock(&product_list_mutex);
    if (dsd_product_list == NULL) {
        err = DsmApiInitializeProductType(eDSD);
        if (err) {
            goto done;
        }

    }
    if (dsm_product_list == NULL) {
        err = DsmApiInitializeProductType(eDSM);
        if (err) {
            goto done;
        }
    }
    if (eth_product_list == NULL) {
        err = DsmApiInitializeProductType(eEthernet);
        if (err) {
            goto done;
        }
    }
done:
    pthread_mutex_unlock(&product_list_mutex);
    return err;
}
